#include "render.h"
#include "program.h"
#include <SDL2/SDL_pixels.h>
#include <SDL2/SDL_render.h>

struct RenderContext* render_init(const struct Program* program, const struct ProgramSettings* settings) {
    struct RenderContext* self = malloc(sizeof(struct RenderContext));

    // create renderer, needs to be able to target textures, preferably hardware accelerated
    self->renderer = SDL_CreateRenderer(program->window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_TARGETTEXTURE);
    // create render target texture
    self->render_target = SDL_CreateTexture(
        self->renderer,
        SDL_PIXELFORMAT_ARGB32,
        SDL_TEXTUREACCESS_TARGET,
        settings->view_resolution.x, settings->view_resolution.y);

    // store the desired view resolution
    self->render_resolution = settings->view_resolution;

    // get the destination area for rendering the target onto the window
    render_calculate_render_area(self);

    // clear target black
    SDL_SetRenderTarget(self->renderer, self->render_target);
    SDL_SetRenderDrawColor(self->renderer, 0, 0, 0, 255);
    SDL_RenderClear(self->renderer);

    return self;
}

void render_destroy(struct RenderContext* self) {
    // clear up allocated SDL objects
    SDL_DestroyTexture(self->render_target);
    SDL_DestroyRenderer(self->renderer);
}

void render_present(struct RenderContext* self) {
    // clear window white
    SDL_SetRenderTarget(self->renderer, NULL);
    SDL_SetRenderDrawColor(self->renderer, 0, 0, 0, 255);
    SDL_RenderClear(self->renderer);
    // copy render target
    SDL_Rect source_rect = {0, 0, self->render_resolution.x, self->render_resolution.y};
    SDL_RenderCopy(self->renderer, self->render_target, &source_rect, &self->render_area);
    SDL_RenderPresent(self->renderer);
    // clear render target
    SDL_SetRenderTarget(self->renderer, self->render_target);
    SDL_SetRenderDrawColor(self->renderer, 0, 0, 0, 255);
    SDL_RenderClear(self->renderer);
}

void render_calculate_render_area(struct RenderContext* self) {
    // get aspect ratios of both the window and the rendertexture
    IVector window_resolution = render_get_window_size(self);
    float window_aspect = (float)window_resolution.x / (float)window_resolution.y;
    float target_aspect = (float)self->render_resolution.x / (float)self->render_resolution.y;

    // calculate the largest area that will fit the entire rendertexture into the window space
    self->render_area = (SDL_Rect) {0, 0, window_resolution.x, window_resolution.y};
    if(window_aspect <= target_aspect) {
        self->render_area.h = window_resolution.x / target_aspect;
        self->render_area.y = (window_resolution.y - self->render_area.h) / 2;
    } else {
        self->render_area.w = window_resolution.y * target_aspect;
        self->render_area.x += (window_resolution.x - self->render_area.w) / 2;
    }
}

IVector render_get_window_size(struct RenderContext* self) {
    // select window as target (store target to reset)
    SDL_Texture* target = SDL_GetRenderTarget(self->renderer);
    SDL_SetRenderTarget(self->renderer, NULL);
    // fetch output size (= window size) of renderer
    int window_width, window_height;
    SDL_GetRendererOutputSize(self->renderer, &window_width, &window_height);
    // reset render target
    SDL_SetRenderTarget(self->renderer, target);
    // construct IVector from fetched output size
    return (IVector){window_width, window_height};
}

void render_handle_resize(struct RenderContext* self) {
    render_calculate_render_area(self);
}
